# Copyright (C) 2005 Graham Ashton <ashtong@users.sourceforge.net>
#
# This module is free software, and you may redistribute it and/or modify
# it under the same terms as Python itself, so long as this copyright message
# and disclaimer are retained in their original form.
#
# IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
# SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF
# THIS CODE, EVEN IF THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#
# THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.  THE CODE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS,
# AND THERE IS NO OBLIGATION WHATSOEVER TO PROVIDE MAINTENANCE,
# SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
#
# $Id: netsyslog.py,v 1.9 2005/11/22 16:35:40 ashtong Exp $
#
# Slightly modified by Xiaming Chen, 2015/05/02.

"""
Syslog module of reSyslog

The format of the UDP packets sent by netsyslog adheres closely to
that defined in U{RFC 3164<http://www.ietf.org/rfc/rfc3164.txt>}. Much
of the terminology used in the RFC has been incorporated into the
names of the classes and properties and is used throughout this
documentation.

"""

import os
import socket
import sys
import time

class Facility:
    """Syslog facilities"""
    KERN, USER, MAIL, DAEMON, AUTH, SYSLOG, \
    LPR, NEWS, UUCP, CRON, AUTHPRIV, FTP = range(12)

    LOCAL0, LOCAL1, LOCAL2, LOCAL3, \
    LOCAL4, LOCAL5, LOCAL6, LOCAL7 = range(16, 24)

class Level:
    """Syslog levels"""
    EMERG, ALERT, CRIT, ERR, \
    WARNING, NOTICE, INFO, DEBUG = range(8)

class PRI(object):

    """The PRI part of the packet.

    Though never printed in the output from a syslog server, the PRI
    part is a crucial part of the packet. It encodes both the facility
    and severity of the packet, both of which are defined in terms of
    numerical constants from the standard syslog module.

    See Section 4.1.1 of RFC 3164 for details.

    """

    def __init__(self, facility, severity):
        """Initialise the object, specifying facility and severity.

        Specify the arguments using constants from the syslog module
        (e.g. syslog.LOG_USER, syslog.LOG_INFO).

        """
        assert facility is not None
        assert severity is not None
        self.facility = facility
        self.severity = severity

    def __str__(self):
        value = self.facility + self.severity
        return "<%s>" % value


class HEADER(object):

    """The HEADER part of the message.

    The HEADER contains a timestamp and a hostname. It is the first
    component of the log message that is displayed in the output of
    syslog.

    See Section 4.1.2 of RFC 3164 for details.

    """

    def __init__(self, timestamp=None, hostname=None):
        """Initialise the object, specifying timestamp and hostname.

        The timestamp represents the local time when the log message
        was created. If the timestamp is not set the current local
        time will be used. See the L{HEADER.timestamp} property
        for a note on the format.

        The hostname should be set to the hostname of the computer
        that originally generated the log message. If the hostname is
        not set the hostname of the local computer will be used. See
        the L{HEADER.hostname} property for a note on the format.

        """
        self.timestamp = timestamp
        self.hostname = hostname

    def __str__(self):
        return "%s %s" % (self.timestamp, self.hostname)

    @property
    def timestamp(self):
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        """
        The local time when the message was written.

        Must follow the format 'Mmm DD HH:MM:SS'.  If
        the day of the month is less than 10, then it
        MUST be represented as a space and then the
        number.

        """
        if not self._timestamp_is_valid(value):
            value = self._calculate_current_timestamp()
        self._timestamp = value

    def _calculate_current_timestamp(self):
        localtime = time.localtime()
        day = time.strftime("%d", localtime)
        if day[0] == "0":
            day = " " + day[1:]
        value = time.strftime("%b %%s %H:%M:%S", localtime)
        return value % day

    def _timestamp_is_valid(self, value):
        if value is None:
            return False
        for char in value:
            if ord(char) < 32 or ord(char) > 126:
                return False
        return True

    @property
    def hostname(self):
        return self._hostname

    @hostname.setter
    def hostname(self, value):
        """
        The hostname where the log message was created.

        Should be the first part of the hostname, or
        an IP address. Should NOT be set to a fully
        qualified domain name.

        """
        if value is None:
            value = socket.gethostname()
        self._hostname = value

class MSG(object):
    """Represents the MSG part of a syslog packet.

    The MSG part of the packet consists of the TAG and CONTENT. The
    TAG and the CONTENT fields must be separated by a non-alphanumeric
    character. Unless you ensure that the CONTENT field begins with
    such a character a seperator of a colon and space will be inserted
    between them when the C{MSG} object is converted into a UDP
    packet.

    See Section 4.1.3 of RFC 3164 for details.

    """

    MAX_TAG_LEN = 32

    def __init__(self, tag=None, content="", pid=None):
        """Initialise the object, specifying tag and content.

        See the documentation for the L{MSG.tag} and
        L{MSG.content} properties for further documentation.

        If the pid is set it will be prepended to the content in
        square brackets when the packet is created.

        """
        self.tag = tag
        self.content = content
        self.pid = pid

    def __str__(self):
        content = self.content
        if self.pid is not None:
            content = "[%s]" % self.pid + content
        return self.tag + content

    @property
    def tag(self):
        return self._tag

    @tag.setter
    def tag(self, value):
        """The name of the program that generated the log message.

        The tag can only contain alphanumeric
        characters. If the tag is longer than {MAX_TAG_LEN} characters
        it will be truncated automatically.

        """
        if value is None:
            value = sys.argv[0]
        self._tag = value[:self.MAX_TAG_LEN]

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, value):
        """The main component of the log message.

        The content field is a freeform field that
        often begins with the process ID (pid) of the
        program that created the message.

        """
        value = self._prepend_seperator(value)
        self._content = value

    def _prepend_seperator(self, value):
        try:
            first_char = value[0]
        except IndexError:
            pass
        else:
            if first_char.isalnum():
                value = ": " + value
        return value

class Packet(object):
    """Combines the PRI, HEADER and MSG into a packet.

    If the packet is longer than L{MAX_LEN} bytes in length it is
    automatically truncated prior to sending; any extraneous bytes are
    lost.

    """

    MAX_LEN = 1024

    def __init__(self, pri, header, msg):
        """Initialise the object.

        The three arguments must be instances of the L{PRI},
        L{HEADER} and L{MSG} classes.

        """
        self.pri = pri
        self.header = header
        self.msg = msg

    def __str__(self):
        message = "%s%s %s" % (self.pri, self.header, self.msg)
        return message[:self.MAX_LEN]


class Syslog(object):
    """Send log messages to syslog servers.

    The Syslog class provides two different methods for sending log
    messages. The first approach (the L{log} method) is suitable for
    creating new log messages from within a normal application. The
    second (the L{send_packet} method) is designed for use in
    circumstances where you need full control over the contents of
    the syslog packet.

    """

    PORT = 514

    def __init__(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._hostnames = {}

    def add_host(self, hostname):
        """Add hostname to the list of hosts that will receive packets.

        Can be a hostname or an IP address. Note that if the hostname
        cannot be resolved calls to L{log} or L{send_packet} will take
        a long time to return.
        """
        self._hostnames[hostname] = 1

    def remove_host(self, hostname):
        """Remove hostname from the list of hosts that will receive packets."""
        del self._hostnames[hostname]

    def host_number(self):
        return len(self._hostnames)

    def _send_packet_to_hosts(self, packet):
        for hostname in self._hostnames:
            host = hostname
            port = self.PORT
            if ':' in hostname:
                host, port = hostname.split(':')
            self._sock.sendto(str(packet), (host, int(port)))

    def log(self, facility, level, text, pid=False):
        """Send the message text to all registered hosts.

        The facility and level will be used to create the packet's PRI
        part. The HEADER will be automatically determined from the
        current time and hostname. The MSG will be set from the
        running program's name and the text parameter.

        This is the simplest way to use reSyslog.Syslog, creating log
        messages containing the current time, hostname, program name,
        etc. This is how you do it::

            logger = syslog.Syslog()
            logger.add_host("localhost")
            logger.log(Facility.USER, Level.INFO, "Hello World")

        If pid is True the process ID will be prepended to the text
        parameter, enclosed in square brackets and followed by a
        colon.

        """
        pri = PRI(facility, level)
        header = HEADER()
        if pid:
            msg = MSG(content=text, pid=os.getpid())
        else:
            msg = MSG(content=text)
        packet = Packet(pri, header, msg)
        self._send_packet_to_hosts(packet)

    def send_packet(self, packet):
        """Send a L{Packet} object to all registered hosts.

        This method requires more effort than L{log} as you need to
        construct your own L{Packet} object beforehand, but it does
        give you full control over the contents of the packet::

            pri = syslog.PRI(Facility.USER, Level.INFO)
            header = syslog.HEADER("Jun  1 18:34:03", "myhost")
            msg = syslog.MSG("myprog", "Hello World", mypid)
            packet = syslog.Packet(pri, header, msg)

            logger = syslog.Syslog()
            logger.add_host("localhost")
            logger.send_packet(packet)

        """
        self._send_packet_to_hosts(packet)



"""
Configuration parser for relogger.

# Quick Start

The parser supports both CLI parameters and configuration file.  As the role of
relogger being a relay of syslog, CLI parameters or the configuration file
mainly aim to supply the source and destination description.

## Using configuration file

    $ relogger -F config_file.txt

The configuration file is consistent with the style of
[RFC 822](https://www.ietf.org/rfc/rfc0822.txt), and can be parsed with
`ConfigParser` module in Python 2.  The configuration of relogger consists of
several [section]s and each section defines a set of rules about sources and
destinations.  A quick example of `config_file.txt` is like:

    [rule1]
    src.port = 514
    dst.host = localhost:666
    dst.file = output.dat

    [rule2]
    src.file = 234,514
    dst.host = 10.50.200.100:6666

The section name is user-defined. Options in each section currently support
`src.port`, `src.file`, `dst.host`, `dst.file`.  In a section, at least one
src-dst pair should be configured.  For port and host descriptors, multiple
values are separated by commas.

## Using command lines

The CLI parameters behaves in a similar way with the configuration file, and
obey the rules described above. Here we give some quick examples:

* Replicate syslog from port 514 to two local ports:

        $ relogger -s localhost:514 -d localhost:30514,localhost:31514

* Reseive syslog from port 514 and save to a file:

        $ relogger -s localhost:514 -w syslog.txt

* Replay an offline file to remote server on port 514:

        $ relogger -r syslog.txt -d 10.50.200.100

* Replicate syslog from port 514 to remote host and offline file
  simultaneously:

        $ relogger -s localhost -d 10.50.200.100:514 -w syslog.txt

By Xiaming Chen <chenxm35@gmail.com>, 2015/05/02.

"""
import re
import os
import ConfigParser

class conferr(Exception): pass

def valid_hostname(hostname):
    validIP = r"^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$"
    validHN = r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$"

    ipregex = re.compile(validIP)
    hnregex = re.compile(validHN)
    return ipregex.match(hostname) or hnregex.match(hostname)

class RLConfig(object):
    """ The general configuration for Relogger

    Quick example:
        config = relogger.RLConfig(config=your_config_file.txt)
        config.flowtable

    """
    SPREFIX = 'src'
    DPREFIX = 'dst'
    HOST = 'localhost'
    PORT = 514

    def __init__(self, source=None, ifile=None, dest=None, ofile=None, config=None):
        self.flow_table = list()

        if config:
            # parse config file
            config_parser = ConfigParser.SafeConfigParser()
            fp = config if isinstance(config, file) else open(config)
            self.config_file = fp.name
            config_parser.readfp(fp)
            sections = config_parser.sections()

            for s in sections:
                fields = self._get_section_values(config_parser, s)
                thistable = self._assemble_flowtable(fields)
                self.flow_table.append(thistable)

        else:
            src_host = self._get_hosts_from_names(source) if source else None
            src_file = ['file://' + os.path.abspath(ifile)] if ifile else None
            dst_host = self._get_hosts_from_names(dest) if dest else None
            dst_file = ['file://' + os.path.abspath(ofile)] if ofile else None

            if src_host is None and src_file is None:
                raise conferr('No source detected, at least one required.')
            if dst_host is None and dst_file is None:
                raise conferr('No destination detected, at least one required.')

            fields = (src_host, src_file, dst_host, dst_file)
            thistable = self._assemble_flowtable(fields)
            self.flow_table.append(thistable)

        self._detect_loop()

    def _get_section_values(self, config, section):
        """ extract src and dst values from a section
        """
        src_host = self._get_hosts_from_ports(config.get(section, 'src.port')) \
            if config.has_option(section, 'src.port') else None
        src_file = [self._get_abs_filepath(config.get(section, 'src.file'))] \
            if config.has_option(section, 'src.file') else None
        if src_host is None and src_file is None:
            raise conferr('Section "%s" gets no sources' % section)

        dst_host = self._get_hosts_from_names(config.get(section, 'dst.host')) \
            if config.has_option(section, 'dst.host') else None
        dst_file = [self._get_abs_filepath(config.get(section, 'dst.file'))] \
            if config.has_option(section, 'dst.file') else None
        if dst_host is None and dst_file is None:
            raise conferr('Section "%s" gets no destinations' % section)

        return (src_host, src_file, dst_host, dst_file)

    def _assemble_flowtable(self, values):
        """ generate a flowtable from a tuple of descriptors.
        """
        values = map(lambda x: [] if x is None else x, values)
        src = values[0] + values[1]
        dst = values[2] + values[3]

        thistable = dict()
        for s in src:
            thistable[s] = dst
        return thistable

    def _detect_loop(self):
        """ detect loops in flow table, raise error if being present
        """
        for source, dests in self.flowtable.items():
            if source in dests:
                raise conferr('Loops detected: %s --> %s' % (source, source))

    def _get_hosts_from_ports(self, ports):
        """ validate hostnames from a list of ports
        """
        hosts = map(lambda x: 'localhost:%d' % int(x.strip()), ports.split(','))
        return list(set(hosts))

    def _get_hosts_from_names(self, names):
        """ validate hostnames from a list of names
        """
        result = set()
        hosts = map(lambda x: x.strip(), names.split(','))
        for h in hosts:
            if valid_hostname(h.split(':')[0]):
                result.add(h if ':' in h else '%s:%d' % (h, self.PORT))
            else:
                raise conferr('Invalid hostname: %s' % h.split(':')[0])
        return list(result)

    def _get_abs_filepath(self, ifile):
        """ validate src or dst file path with self.config_file
        """
        assert ifile is not None
        ifile = ifile[7:] if ifile.startswith('file://') else ifile
        if ifile[0] != '/':
            basedir = os.path.abspath(os.path.dirname(self.config_file))
            ifile = os.path.join(basedir, ifile)
        return 'file://' + ifile

    @property
    def flowtable(self):
        """ get a flat flow table globally
        """
        ftable = dict()
        for table in self.flow_table:
            for k, v in table.items():
                if k not in ftable:
                    ftable[k] = set(v)
                else:
                    [ftable[k].add(i) for i in v]
        # convert set to list
        for k in ftable:
            ftable[k] = list(ftable[k])
        return ftable

    @property
    def flowtables(self):
        """ get a list of flow table for individual source-dest pairs
        """
        return self.flow_table

    def has_source_socket(self):
        hosts = [ i for i in self.flowtable if not i.startswith('file://') ]
        return len(hosts) > 0

    def has_source_file(self):
        files = [ i for i in self.flowtable if i.startswith('file://') ]
        return len(files) > 0


"""
Relogger server to read UDP logs from multiple sources.
"""
import threading
import SocketServer
from Queue import Queue

from syslog import Syslog

class RLServer(object):

    def __init__(self, flowtable):
        self.flowtable = flowtable
        self.message_queue = Queue()

    @property
    def flowtable(self):
        return self._flowtable

    @flowtable.setter
    def flowtable(self, value):
        self._flowtable = {}
        for k, v in value.items():
            ofiles = []
            logger = Syslog()
            for i in v:
                if i.startswith('file://'):
                    ofiles.append(open(i.replace('file://', ''), 'wb'))
                else:
                    logger.add_host(i)
            self._flowtable[k] = (logger, ofiles)

    def _serve_socket(self, host, port):
    	""" utility function to read socket and send to destinations
    	"""
        mqueue = self.message_queue

        class UDPHandler(SocketServer.BaseRequestHandler):
            def handle(self):
                data = self.request[0].strip()
                mqueue.put(('%s:%d' % (host, port), data))

        server = SocketServer.UDPServer((host, port), UDPHandler)
        server.serve_forever()

    def _serve_file(self, filename, count):
    	""" utility function to read file and send to destinations
    	"""
        for line in open(filename.replace('file://', ''), 'rb'):
            data = line.strip(' \r\n')
            self.message_queue.put((filename, data))

    def _message_consumer(self):
        while True:
            source, data = self.message_queue.get()
            logger, ofiles = self.flowtable[source]
            # sending message
            if logger.host_number > 0:
                logger.send_packet(data)
            if len(ofiles) > 0:
                for f in ofiles:
                    f.write(data + '\n')
                    f.flush()
            self.message_queue.task_done()

    def start(self):
    	"""
    	A quick example:
    	    rlconfig = RLConfig(config=args.config)
    	    server = RLServer(rlconfig.flowtable)
    	    server.start()
    	"""
    	thread_pool = []

        # message consumer
        t = threading.Thread(target=self._message_consumer)
        t.setDaemon(True)
        thread_pool.append(t)

        # message producers
        for source, dest in self._flowtable.items():
            if not source.startswith('file://'):
                host, port = source.split(':')
                t = threading.Thread(target=self._serve_socket, args=(host, int(port)))
                t.setDaemon(True)
                thread_pool.append(t)
            else:
            	t = threading.Thread(target=self._serve_file, args=(source, -1))
            	t.setDaemon(True)
            	thread_pool.append(t)

        # start all threads
        [ t.start() for t in thread_pool ]


import os
import sys
from setuptools import setup
from relogger import __version__

setup(
    name = "relogger",
    version = __version__,
    url = 'https://github.com/caesar0301/relogger',
    author = 'Xiaming Chen',
    author_email = 'chenxm35@gmail.com',
    description = 'A syslog sender, relay and receiver.',
    long_description='''A relayer or replicator to send SYSLOG
from one or multiple sources to one or multiple destinations.''',
    license = "The MIT License",
    packages = ['relogger'],
    scripts=['bin/relogger'],
    keywords = ['syslog', 'relay'],
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
            'Intended Audience :: Developers',
            'License :: Freely Distributable',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Topic :: Software Development :: Libraries :: Python Modules',
   ],
)